%% Implementation of the single shooting scheme of optimal control problems
%close all
clear variables
clc

%% Create NLP
import casadi.*

N   = 30;       % Control discretization
T   = 3.0;      % End time
h   = T/N;      % Discretization step
nx  = 1;            
nu  = 1;

Q = 1;
R = 1;

% Declare variables (use scalar graph)
u  = SX.sym('u',nu);    % control
x  = SX.sym('x',nx);    % states

% System dynamics
xdotl= -x(1)  + u(1);
xdotr= -(1 + x(1))*(x(1))  + u(1);
f = Function('f', {x,u},{xdotl});
fr = Function('fr', {x,u},{xdotr});

% RK4
U   = MX.sym('U');
X0  = MX.sym('X0',nx);
X   = X0;

k = f({X,U});
k1 = k{1};
k = f({X + h/2 * k1, U});
k2 = k{1};
k = f({X + h/2 * k2, U});
k3 = k{1};
k = f({X + h   * k3, U});
k4 = k{1};
X = X + h/6*(k1   + 2*k2   + 2*k3   + k4);
RK4 = Function('RK4',{X0,U},{X});

X = X0;
k = fr({X,U});
k1 = k{1};
k = fr({X + h/2 * k1, U});
k2 = k{1};
k = fr({X + h/2 * k2, U});
k3 = k{1};
k = fr({X + h   * k3, U});
k4 = k{1};
X = X + h/6*(k1   + 2*k2   + 2*k3   + k4);
RK4r = Function('RK4r',{X0,U},{X});

% Formulate NLP 
u = SX.sym('u',N);

% Objective function
J=0;

% Get an expression for the cost and state at end

x_0 = 1;
X = x_0;
g = []; lbg = []; ubg = [];
min_x = 0.5;
for i = 1:N
    out = RK4({X,u(i)});
    X = out{1};
    if (i >= 15 && i <= 17)
        g = [g; X];
        ubg = [ubg; Inf];
        lbg = [lbg; min_x];
    end
    J = J + Q*X(1)^2 + R*u(i)^2;
end

% Terminal constraints: x_0(T)=x_1(T)=0
g = [g;X];
ubg = [ubg; 0];
lbg = [lbg; 0];
% Allocate an NLP solver
nlp = struct('x', u, 'f', J, 'g', g);

% Create IPOPT solver object
options = struct;
options.hessian_approximation = 'limited-memory';
solver = nlpsol('solver', 'ipopt', nlp);

%% Solve the NLP
arg = struct;
arg.x0  = 0.;    % solution guess
arg.lbx = -2;    % lower bound on x
arg.ubx =  2;    % upper bound on x
arg.lbg =  lbg;    % lower bound on g
arg.ubg =  ubg;    % upper bound on g

% Solve the problem
res = solver(arg);

f_opt       = full(res.f);
u_opt       = full(res.x);

% Compute state trajectory integrating the dynamics
x_opt       = zeros(nx,N+1);
x_optr      = zeros(nx,N+1);
x_lqr       = zeros(nx,N+1);
x_opt(1)    = x_0;
x_optr(1)   = x_0;
x_lqr(1)    = x_0;

for i=2:N+1
    out         = RK4({x_opt(i-1),u_opt(i-1)});
    x_opt(i)    = full(out{1});
    out         = RK4r({x_optr(i-1),u_opt(i-1)});
    x_optr(i)   = full(out{1});
    out         = RK4r({x_lqr(i-1),-0.0501*x_lqr(i-1)});
    x_lqr(i) = full(out{1});
end

%% Plot results

figure()
subplot(2,1,1)
plot((0:N)*T/N, x_opt,(0:N)*T/N, x_optr,(0:N)*T/N, x_lqr)
xlabel('time - t')
ylabel('state - x')
grid on
hold all
plot((15:17)*h, min_x*ones(1,3),'-','LineWidth',1)
legend('exact','linearized','lqr')
subplot(2,1,2)
stairs((0:N-1)*T/N, u_opt,'r')
xlabel('time - t')
ylabel('input - u')
grid on

